#include <avr32/io.h>
#include "macro.h"


.global dataflash_init
.global dataflash_readpage
.global dataflash_readmanufacturer
.global dataflash_waitRecv
.global dataflash_writepage


dataflash_init:
    stm     --sp,r8,r9,lr

    lddpc   r8,.spi0

    /* reset SPI0 */
    mov     r9,0b10000000
    st.w    r8[AVR32_SPI_CR],r9

    /* set Master, use CS0, MCK = PBA/1 = 150MHZ*/
    mov     r9,0x0011
    st.w    r8[AVR32_SPI_MR],r9

    /* configure chip select */
    mov     r9,0x0109                   // PBA=main/4=37.5MHZ. spi=pba/1 , stay active. Atmel suggest going no more than 20MHZ
    orh     r9,0x0A00                   // delay of 10 between each byte.
    st.w    r8[AVR32_SPI_CSR0],r9


    // Assign pins to SPI
    mov     r9,(1<<AVR32_SPI0_SCK_0_PIN)|(1<<AVR32_SPI0_MISO_0_PIN)|(1<<AVR32_SPI0_MOSI_0_PIN)|(1<<AVR32_SPI0_NPCS_0_PIN)
    lddpc   r7,.pioa

    st.w    r7[AVR32_PIO_PDR],r9        // disable GPIO control
    st.w    r7[AVR32_PIO_ASR],r9        // peripheral A
    st.w    r7[AVR32_PIO_PUDR],r9       // pull-up disable


    /* enable SPI0 */
    mov     r9,1
    st.w    r8[AVR32_SPI_CR],r9

    ldm     sp++,r8,r9,pc

.pioa: .int AVR32_PIOA_ADDRESS
.spi0:  .int 0xFFE00000


//////////////////////////////////////////////
/// write SPI
/// INPUT: r6=byte to send
/// OUTPUT: r12=byte that was shifted in during write
/////////////////////////////////////////////
writeSPI:
    stm     --sp,r7,r9,lr

    lddpc   r9,.spi0
    st.w    r9[AVR32_SPI_TDR],r6

1:  ld.w    r7,r9[AVR32_SPI_SR]
    bld     r7,AVR32_SPI_TXEMPTY_OFFSET
    brne    1b

    ld.w    r12,r9[AVR32_SPI_RDR]
    ldm     sp++,r7,r9,pc




//////////////////////////////////////////////
/// write dataflash_readpage
/// INPUT: r8=buffer to receive page data. must be in P1 or P2
/// INPUT: r7=page Number
/// INPUT: r5=0 for blocking, non-zero if not blocking
/////////////////////////////////////////////
dataflash_readpage:
    stm     --sp,r6,r9,lr

    /* wait for dataflash to be ready */
    rcall   dataflash_waitready

    /* send opcode */
    mov     r6,0xD2             // main memory page read
    rcall   writeSPI

    /* send address */
    bfextu  r6,r7,5,8           // 8 MSB of page
    rcall   writeSPI
    mov     r6,0
    bfins   r6,r7,3,5           // 5 LSB of page
    rcall   writeSPI
    mov     r6,0                // offset 0
    rcall   writeSPI

    
    /* send 4 dummy writes*/
    mov     r9,4
1:  rcall   writeSPI
    sub     r9,1
    brne    1b

    /* start PDC */
    lddpc   r9,.spi0
    mov     r7,0x0202               // disable transfer
    st.w    r9[0x120],r7
    mov     r7,1056
    st.w    r9[0x100],r8
    st.w    r9[0x104],r7
    mov     r8,0                    // read from anywhere, these are dummy reads
    st.w    r9[0x108],r8
    st.w    r9[0x10C],r7
    mov     r7,0x0101               // enable transfers
    st.w    r9[0x120],r7


    /* Should we wait for completion? */
    cp      r5,0
    brne    noWaitPDC
    rcall   dataflash_waitRecv
noWaitPDC:
    /*WARNING: when we are not toggling CS back after transfer. But it is working... */
    ldm     sp++,r6,r9,pc



//////////////////////////////////////////////
/// dataflash_readmanufacturer
/////////////////////////////////////////////
dataflash_readmanufacturer:
    stm     --sp,r8,r9,lr

    mov     r6,0x9F
    rcall   writeSPI


    rcall   writeSPI // send dummy byte
    mov     r8,r12
    rcall   debug_showRegister

    rcall   writeSPI // send dummy byte
    mov     r8,r12
    rcall   debug_showRegister

    rcall   writeSPI // send dummy byte
    mov     r8,r12
    rcall   debug_showRegister

    sbr     r6,24                                   // end of transmission                         
    rcall   writeSPI // send dummy byte
    mov     r8,r12
    rcall   debug_showRegister

    SERIALPRINTF("\r\n")
    ldm     sp++,r8,r9,pc


/******************************************************************
*** FUNCTION: dataflash_writepage
*** DESC: This will erase and write a full page to dataflash
*** INPUT: r8=buffer to send. must be in P1 or P2
*** INPUT: r7=page Number to write to
*** INPUT: r5=dont write if page has not changed (0=dont, 1=do it anyways)
*** RETURN: r12=number of bytes copied (0 if page was unchanged and r5==0)
******************************************************************/
dataflash_writepage:
    stm     --sp,r6,r9,lr

    /* wait for dataflash to be ready */
    rcall   dataflash_waitready

    /* opcode */
    mov     r6,0x87     // buffer write
    rcall   writeSPI

    /* 13 dont care bits and 11 buffer index bits*/
    mov     r6,0
    rcall   writeSPI
    mov     r6,0
    rcall   writeSPI
    mov     r6,0
    rcall   writeSPI

    /* start PDC to upload*/
    lddpc   r9,.spi0
    mov     r6,0x0202               // disable transfer5
    st.w    r9[0x120],r6
    mov     r6,1056
    st.w    r9[0x108],r8
    st.w    r9[0x10C],r6
    mov     r6,0x0100               // enable transfers
    st.w    r9[0x120],r6
    
    /* wait for PDC */
1:  ld.w    r6,r9[AVR32_SPI_SR]
    bld     r6,7                    // if TXBUFE == 1, PDC is finished
    brne    1b

    cp      r5,0
    brne    skipCompare

    /* now compare the buffer to the page in flash */
    mov     r6,0x61     // buffer compare
    rcall   writeSPI
    
    /* send address of page to compare */
    bfextu  r6,r7,5,8           // 8 MSB of page
    rcall   writeSPI
    mov     r6,0
    bfins   r6,r7,3,5           // 5 LSB of page
    rcall   writeSPI
    mov     r6,0                // offset 0
    sbr     r6,24               // LASTXFER
    rcall   writeSPI

    /* get compare result */
    mov     r6,0xD7     // buffer compare
    rcall   writeSPI
1:  rcall   writeSPI    // dummy write
    bld     r12,7
    brne    1b          // Z need to be 1 to indicate device not busy
    sbr     r6,24       // send another dummy byte and a LASTXFER bit
    rcall   writeSPI    // get the page match flag
    bld     r12,6
    brne    pageAreSame     // 0 = page match

skipCompare:
    
    /* write buffer with built-in erase*/
    mov     r6,0x86             // buffer write with built-in erase
    rcall   writeSPI

    /* send address of page to write */
    bfextu  r6,r7,5,8           // 8 MSB of page
    rcall   writeSPI
    mov     r6,0
    bfins   r6,r7,3,5           // 5 LSB of page
    rcall   writeSPI
    mov     r6,0                // offset 0
    sbr     r6,24               // LASTXFER
    rcall   writeSPI

    /* We return now without waiting to see if chip is busy. We will check before next operation instead */
    mov     r12,1056
    rjmp    2f

pageAreSame:
    mov     r12,0
2:
    ldm     sp++,r6,r9,pc


/******************************************************************
** DESC: wait for chip to become ready
******************************************************************/
dataflash_waitready:
    stm     --sp,r6,r12,lr

1:  mov     r6,0xD7     // buffer compare
    rcall   writeSPI
    sbr     r6,24
    rcall   writeSPI    // dummy write
    bld     r12,7
    brne    1b          // Z need to be 1 to indicate device not busy

    ldm     sp++,r6,r12,pc


/******************************************************************
** DESC: Wait for PDC to be done.
** DESC: Usefull when using dataflash_readpage in non-blocking mode
******************************************************************/
dataflash_waitRecv:
    stm     --sp,r7,r9,lr

    /* wait for PDC to be done */
    lddpc   r9,.spi0
1:  ld.w    r7,r9[AVR32_SPI_SR]
    bld     r7,6
    brne    1b

    ldm     sp++,r7,r9,pc


