#include <avr32/io.h>
#include "macro.h"

/*************************************************
** Unfortunately, it is impossible to detect
** link status trough an interrupt. This would
** be done by configuring the PHY chip
** to generate an interrupt on an output pin
** on link status change, and detect that pin change
** on the AVR32 chip. But the PHY pin
** is not connected on the AVR32 chip.
*************************************************/

#define RX_BUF_COUNT            48
#define TX_BUF_COUNT            12
#define RX_BUF_SIZE             128
#define TX_BUF_SIZE             512
#define MACB0_BUF_BASE          0xB1FC0000
#define MACB1_BUF_BASE          0xB1FC3200
#define MACB_RX_TABLE           0
#define MACB_TX_TABLE           0x180
#define MACB_RX_BUF             0x1E0
#define MACB_TX_BUF             0x19E0
#define MACB_CURRENT_RX         0x31E0
#define MACB_CURRENT_TX         0x31E4


#define MACSTRUCT_MACB  0
#define MACSTRUCT_PIORMII 4
#define MACSTRUCT_PINSRMII 8
#define MACSTRUCT_MACADD 12
#define MACSTRUCT_BUFBASE 20
#define MACSTRUCT_PHYID 24
#define MACSTRUCT_PIOMII 32
#define MACSTRUCT_PINSMII 28

.global init_macb
.global macb_getframe
.global macb_sendframe

/******************************************
** r11 = MACB struct
** r9 = address
** RETURN: r12 = value
******************************************/
mdio_read:
    stm     --sp,r5-r9,lr

    ld.w    r8,r11[MACSTRUCT_MACB]
    ld.w    r7,r8[AVR32_MACB_NCR]
    sbr     r7,AVR32_MACB_NCR_MPE
    st.w    r8[AVR32_MACB_NCR],r7
    
    lddpc   r7,.man
    or      r7,r7,r9<<AVR32_MACB_REGA_OFFSET
    ld.w    r9,r11[MACSTRUCT_PHYID]
    or      r7,r7,r9<<AVR32_MACB_PHYA_OFFSET
    st.w    r8[AVR32_MACB_MAN],r7
1:  ld.w    r7,r8[AVR32_MACB_NSR]
    bld     r7,AVR32_MACB_NSR_IDLE
    brne    1b
    
    ld.w    r12,r8[AVR32_MACB_MAN]
    andh    r12,0

    ld.w    r7,r8[AVR32_MACB_NCR]
    cbr     r7,AVR32_MACB_NCR_MPE
    st.w    r8[AVR32_MACB_NCR],r7

    ldm     sp++,r5-r9,pc
.man: .int (1<<AVR32_MACB_SOF)|(2<<AVR32_MACB_CODE_OFFSET)|(2<<AVR32_MACB_RW_OFFSET)

/******************************************
** r11 = MACB struct
** r9 = address
** r7 = data
******************************************/
mdio_write:
    stm     --sp,r5-r9,lr

    ld.w    r8,r11[MACSTRUCT_MACB]
    ld.w    r6,r8[AVR32_MACB_NCR]
    sbr     r6,AVR32_MACB_NCR_MPE
    st.w    r8[AVR32_MACB_NCR],r6

    lddpc   r6,.manwrite
    or      r6,r6,r9<<AVR32_MACB_REGA_OFFSET
    ld.w    r9,r11[MACSTRUCT_PHYID]
    or      r6,r6,r9<<AVR32_MACB_PHYA_OFFSET
    or      r7,r6
    st.w    r8[AVR32_MACB_MAN],r7

1:  ld.w    r7,r8[AVR32_MACB_NSR]
    bld     r7,AVR32_MACB_NSR_IDLE
    brne    1b

    ld.w    r12,r8[AVR32_MACB_MAN]
    andh    r12,0

    ld.w    r7,r8[AVR32_MACB_NCR]
    cbr     r7,AVR32_MACB_NCR_MPE
    st.w    r8[AVR32_MACB_NCR],r7

    ldm     sp++,r5-r9,pc
.manwrite: .int (1<<AVR32_MACB_SOF)|(2<<AVR32_MACB_CODE_OFFSET)|(1<<AVR32_MACB_RW_OFFSET)



/******************************************
** r11 = MACB struct
** RETURN: r12 = 0 if sucess
******************************************/
init_phy:
    stm     --sp,r5-r9,lr

    /* start auto-negotiation but advertise 100mbps FD only. */
    SERIALPRINTF("Starting auto-negotiation\r\n")
    mov     r9,4                            // ANAR
    mov     r7,0b100000001                  // advertise that we only support 100 full duplex
    rcall   mdio_write

    mov     r9,0                            // BMCR
    mov     r7,0b1001000000000              // enable auto-neg and restart
    rcall   mdio_write

    /* wait for negotiation to be completed */
    mov     r5,0x80000
1:  sub     r5,1
    breq    init_phy_exit_err
    mov     r9,1                            // BMSR
    rcall   mdio_read
    bld     r12,5                           // check auto-neg status     
    brne    1b
    SERIALPRINTF("Negotiation complete. Waiting for link status\r\n")

    /* wait for link status to be up */
1:  mov     r9,1                            // BMSR
    rcall   mdio_read
    bld     r12,2                           // check link status
    brne    1b

    /* does peer support 100 FD? */
    bld     r12,14
    brne    init_phy_exit_err

    SERIALPRINTF("Link status is UP\r\n")
    rjmp    init_phy_exit
init_phy_exit_err:
    SERIALPRINTF("Link status is DOWN\r\n")

init_phy_exit:

    //mov r8,0
    //1:mov   r9,1                            // BMSR
    //rcall   mdio_read
    //cp r12,r8
    //breq 1b
    //SERIALPRINTF("\r\nNew Status: ")
    //mov r8,r12
    //rcall debug_showRegister
    //rjmp    1b

    ldm     sp++,r5-r9,pc

/***************************************************************
** Initiate MACB. 100mbps full duplex using MII
** r5 = MACB number
***************************************************************/
init_macb:
    stm     --sp,r5-r9,r11,lr

    cp      r5,0
    brne    1f
    mov     r11,LO(.MACB0)
    orh     r11,HI(.MACB0)
    rjmp    2f
1:  mov     r11,LO(.MACB1)
    orh     r11,HI(.MACB1)
2:  

     /* setup GPIO (RMII only)*/
    ld.w    r9,r11[MACSTRUCT_PINSRMII]
    ld.w    r8,r11[MACSTRUCT_PIORMII]
    st.w    r8[AVR32_PIO_PDR],r9            // disable GPIO control
    cp      r5,0
    brne    1f
    st.w    r8[AVR32_PIO_ASR],r9            // peripheral A
    rjmp    2f
1:  st.w    r8[AVR32_PIO_BSR],r9            // peripheral B
2:  st.w    r8[AVR32_PIO_PUDR],r9           // pull-up disable

    /*setup GPIO for extra pins needed by MII */
    ld.w    r9,r11[MACSTRUCT_PINSMII]
    ld.w    r8,r11[MACSTRUCT_PIOMII]
    st.w    r8[AVR32_PIO_PDR],r9            // disable GPIO control
    cp      r5,0
    brne    1f
    st.w    r8[AVR32_PIO_ASR],r9            // peripheral A
    rjmp    2f
1:  st.w    r8[AVR32_PIO_BSR],r9            // peripheral B
2:  st.w    r8[AVR32_PIO_PUDR],r9           // pull-up disable


    /* reset */
    ld.w    r8,r11[MACSTRUCT_MACB]
    mov     r9,0
    st.w    r8[AVR32_MACB_NCR],r9
    com     r9
    st.w    r8[AVR32_MACB_TSR],r9    
    st.w    r8[AVR32_MACB_RSR],r9
    st.w    r8[AVR32_MACB_IDR],r9
    mov     r9,1
    st.w    r8[AVR32_MACB_USRIO],r9         // use MII
    
    /* load mac address */
    ld.w    r9,r11[MACSTRUCT_MACADD+2]
    st.w    r8[AVR32_MACB_SA1B],r9
    ld.uh   r9,r11[MACSTRUCT_MACADD]
    st.w    r8[AVR32_MACB_SA1T],r9

    /* prepare RX descriptors table */
    mov     r7,RX_BUF_COUNT
    ld.w    r8,r11[MACSTRUCT_BUFBASE]                      // MACB buf base
    sub     r8,-(MACB_RX_TABLE)
    ld.w    r9,r11[MACSTRUCT_BUFBASE]                      // MACB buf base
    sub     r9,-(MACB_RX_BUF)
    andh    r9,0x1FFF                                       // load physical addresses in descriptors
    mov     r6,0
1:  cp      r7,1                            // is it the last entry?
    brhi    2f
    orl     r9,0b00000010                   // if last entry, set the "wrap" flag
2:  st.w    r8[0],r9
    st.w    r8[4],r6
    sub     r9,-(RX_BUF_SIZE)
    sub     r8,-8
    sub     r7,1
    brne    1b

    /* prepare TX descriptor table */
    mov     r7,TX_BUF_COUNT
    ld.w    r8,r11[MACSTRUCT_BUFBASE]                      // MACB buf base
    sub     r8,-(MACB_TX_TABLE)
    ld.w    r9,r11[MACSTRUCT_BUFBASE]                      // MACB buf base
    sub     r9,-(MACB_TX_BUF)
    andh    r9,0x1FFF                                       // load physical addresses in descriptors
    mov     r6,0
    sbr     r6,31
1:  cp      r7,1                            // is it the last entry?
    brhi    2f
    orl     r6,0b00000010                   // if last entry, set the "wrap" flag (in the status WORD for Tx descriptor)
2:  st.w    r8[0],r9
    st.w    r8[4],r6
    sub     r9,-(TX_BUF_SIZE)
    sub     r8,-8
    sub     r7,1
    brne    1b

    /* set descriptor table addresses */
    ld.w    r8,r11[MACSTRUCT_MACB]
    ld.w    r9,r11[MACSTRUCT_BUFBASE]                      // MACB buf base
    mov     r7,0
    st.w    r9[MACB_CURRENT_RX],r7
    sub     r9,-(MACB_RX_TABLE)
    andh    r9,0x1FFF                                       // load physical address
    st.w    r8[AVR32_MACB_RBQP],r9
    ld.w    r9,r11[MACSTRUCT_BUFBASE]                      // MACB buf base
    mov     r7,0
    st.w    r9[MACB_CURRENT_TX],r7
    sub     r9,-(MACB_TX_TABLE)
    andh    r9,0x1FFF                                       // load physical address
    st.w    r8[AVR32_MACB_TBQP],r9

    /* setup configuration register */
    lddpc   r9,.ncfgr

    //sbr r9,4    // SETTING TO PROMISCUOUS FOR TESTING PURPOSES
    st.w    r8[AVR32_MACB_NCFGR],r9        

    /* init phy: auto-negotiation, 100mbps full duplex only */
    rcall   init_phy

    /* enable receiver and transmitter */
    mov     r9,(AVR32_MACB_NCR_TE_MASK | AVR32_MACB_NCR_RE_MASK)
    st.w    r8[AVR32_MACB_NCR],r9

    ldm     sp++,r5-r9,r11,pc

/******************************************
** r5 = MACB number
** r9 = buffer from which to send the frame
** r7 = frame size
** RETURN: r12 = number of bytes sent
******************************************/
macb_sendframe:
    stm     --sp,r0-r9,r11,lr

    cp      r5,0
    brne    1f
    mov     r11,LO(.MACB0)
    orh     r11,HI(.MACB0)
    rjmp    2f
1:  mov     r11,LO(.MACB1)
    orh     r11,HI(.MACB1)
2:  ld.w    r8,r11[MACSTRUCT_MACB]

    /* Get current entry value */
    ld.w    r8,r11[MACSTRUCT_BUFBASE]
    ld.w    r1,r8[MACB_CURRENT_TX]                      // r1 = descriptor index
    sub     r0,r8,-(MACB_TX_TABLE)
    sub     r2,r8,-(MACB_TX_BUF)
    mov     r6,r1
    lsl     r6,3
    add     r0,r0,r6                                    // r0 = absolute pointer to descriptor in table
    lsl     r6,4                                        // shift another 4bits, for a total of 7 = mul by 128
    add     r2,r2,r6                                    // r2 = buffer base address

    /* save pointer of first descriptor, will need to set a flag later */
    mov     r4,r0

    /* iterate through all free buffers until required size is found */
nextframetx:
    ld.w    r5,r0[0]
    ld.w    r3,r0[4]
    // TODO fill buffer and decrease r7
    // TODO setup descriptor

    
    bld     r3,1
    breq    txwrap
    sub     r1,-1
    sub     r0,-8
    rjmp    calcnexttx
txwrap:
    sub     r0,r8,-(MACB_TX_TABLE)
    sub     r2,r8,-(MACB_TX_BUF)
    mov     r1,0
calcnexttx:
    cp      r7,0
    brne    nextframetx

    /* set ownership of first descriptor now that we are ready to send */

    // sync all writes

    // enable transmission
    // mark frame size

    ldm     sp++,r0-r9,r11,pc


/******************************************
** r5 = MACB number
** r9 = buffer in which to receive the frame (must be at least 1536 bytes)
** RETURN: r12 = number of bytes received
******************************************/
macb_getframe:
    stm     --sp,r0-r9,r11,lr

    cp      r5,0
    brne    1f
    mov     r11,LO(.MACB0)
    orh     r11,HI(.MACB0)
    rjmp    2f
1:  mov     r11,LO(.MACB1)
    orh     r11,HI(.MACB1)
2:  ld.w    r8,r11[MACSTRUCT_MACB]

    /* Get current entry value */
    ld.w    r8,r11[MACSTRUCT_BUFBASE]
    ld.w    r1,r8[MACB_CURRENT_RX]                      // r1 = descriptor index
    sub     r0,r8,-(MACB_RX_TABLE)
    sub     r2,r8,-(MACB_RX_BUF)
    mov     r7,r1
    lsl     r7,3
    add     r0,r0,r7                                    // r0 = absolute pointer to descriptor in table
    lsl     r7,4                                        // shuft another 4bits, for a total of 7 = mul by 128
    add     r2,r2,r7                                    // r2 = buffer base address

    /* read descriptor */
    cache   r0[0],0b1011                              // need to invalidate cache line. Even if addr is in P2
    cache   r0[4],0b1011                              // need to invalidate cache line. Even if addr is in P2
    ld.w    r7,r0[0]
    ld.w    r3,r0[4]

    /* check ownership */
    bld     r7,0
    brne    macb_getframe_error_noinc                 // we don't have ownership of this buffer

    /* check if SOF */
    bld     r3,14
    brne    macb_getframe_error                       //  if not SOF, this is invalid
    
checkNextFrame:
    /* copy data to user buffer. Dont get address from descriptor since we need the virtual address. */
    mov     r5,32
1:  ld.w    r6,r2++
    st.w    r9++,r6
    sub     r5,1
    brne    1b

    /* give back ownership to MACB */
    cbr     r7,0    
    st.w    r0[0],r7
    sync    0    

    /* should wrap? */
    bld     r7,1
    brne    2f  
    
    /* wrap */
    sub     r0,r8,-(MACB_RX_TABLE)
    sub     r2,r8,-(MACB_RX_BUF)
    mov     r1,0
    rjmp    3f
2:  sub     r0,-8
    sub     r1,-1
3:  bld     r3,15                                       // EOF? 
    breq    end_of_packet
    cache   r0[0],0b1011                              // need to invalidate cache line. Even if addr is in P2
    cache   r0[4],0b1011
    ld.w    r7,r0[0]
    ld.w    r3,r0[4]
    bld     r7,0
    brne    macb_getframe_error_noinc                 // buffer is not ours. return and discard current frame
    bld     r3,14
    breq    macb_getframe_error_noinc                   // SOF but we did not find the EOF before.

    rjmp    checkNextFrame

end_of_packet:
    /* update current RX value */
    st.w    r8[MACB_CURRENT_RX],r1

    /* get frame size */
    mov     r12,r3
    andl    r12,0b111111111111,COH

    rjmp    macb_getframe_exit
macb_getframe_error:
    sub     r1,-1
    cp      r1,RX_BUF_COUNT
    moveq   r1,0
    st.w    r8[MACB_CURRENT_RX],r1                      // increase current pointer to ski pcurrent frame
macb_getframe_error_noinc:
    mov     r12,0
macb_getframe_exit:
    ldm     sp++,r0-r9,r11,pc


/*
AVR32_MACB_NCFGR_CLK = 2 = clk/32. This will be used for MDIO operations only. should result in a clk < 2.5MHZ
*/
.ncfgr:         .int    ((2<<AVR32_MACB_NCFGR_CLK) | (1<<AVR32_MACB_NCFGR_DRFCS)|\
                        (1<<AVR32_MACB_NCFGR_SPD) |(1<<AVR32_MACB_NCFGR_FD))
/******************* MACB0 config ******************************************************/
.MACB0:
.macb0:         .int    AVR32_MACB0_ADDRESS
.pioc:          .int    AVR32_PIOC_ADDRESS
.macb0Pins:     .int    ((1<<AVR32_MACB0_MDC_0_PIN)|(1<<AVR32_MACB0_MDIO_0_PIN)|\
                        (1<<AVR32_MACB0_RXD_0_PIN)|(1<<AVR32_MACB0_TXD_0_PIN)|\
                        (1<<AVR32_MACB0_RXD_1_PIN)|(1<<AVR32_MACB0_TXD_1_PIN)|\
                        (1<<AVR32_MACB0_TX_EN_0_PIN)|(1<<AVR32_MACB0_RX_ER_0_PIN)|\
                        (1<<AVR32_MACB0_RX_DV_0_PIN)|(1<<AVR32_MACB0_TX_CLK_0_PIN))
.macAddress0:   .byte    'P','A','T',2,4,2,0,0
.macb0BufBase:  .int    MACB0_BUF_BASE
.phy0:          .int    1
.mac0PinsMII:   .int    (1<<0)|(1<<1)|(1<<2)|(1<<5)|(1<<6)|(1<<11)|(1<<12)|(1<<14)|(1<<18)
.pioMII0:       .int    AVR32_PIOC_ADDRESS 
/******************* MACB1 config ******************************************************/
.MACB1:
.macb1:         .int    AVR32_MACB1_ADDRESS
.piod:          .int    AVR32_PIOD_ADDRESS
.macb1Pins:     .int    ((1<<3)|(1<<2)|(1<<10)|(1<<13)|(1<<6)|(1<<14)|(1<<11)|(1<<5)|(1<<4)|(1<<12))
.macAddress1:   .byte    'P','A','T',2,4,3,0,0
.macb1BufBase:  .int    MACB1_BUF_BASE
.phy1:          .int    3
.mac1PinsMII:   .int    (1<<19)|(1<<23)|(1<<26)|(1<<27)|(1<<28)|(1<<29)|(1<<30)|(1<<24)|(1<<15)
.pioMII1:       .int    AVR32_PIOC_ADDRESS

