#include <avr32/io.h>
#include "macro.h"

.global mci_init
.global mci_readpage
.global mci_writepage

/**************************************************
** This driver only works for non-HC SD cards.
** No MMC and no SDHC
**
**************************************************/


/**************************************************
** mci_read: read 1 sector
** r8 = buffer to read
** r7 = sector number
***************************************************/
mci_writepage:
    stm     --sp,r0-r6,r9-r11,lr

    /* Check if card inserted and initialized. */
    rcall   sd_init
    cp      r12,0
    brne    write_error

    lsl     r7,9                            // multiply by 512, for non-HC cards, block is address in bytes

    /* wait for ready status */
    lddpc   r6,.mciaddr2
    mov     r11,1000                        // retry 1000 times before assuming failure
1:  sub     r11,1
    breq    write_error
    ld.w    r10,r6[AVR32_MCI_SR]
    bld     r10,5
    brne    1b

    /* CMD13 */
    lddpc   r10,.mmc_rca1
    ld.w    r11,r10[0]
    mov     r10,0x104D
    rcall   mci_cmd


    /* Set block count (1 block of 512 bytes) */
    mov     r9,1
    orh     r9,512
    lddpc   r6,.mciaddr1
    st.w    r6[AVR32_MCI_BLKR],r9
    mov     r10,0x11059                     // multiblock write
    mov     r11,r7
    rcall   mci_cmd

    /* write the data */
    mov     r10,128                         // 128 31bit words
1:  mov     r11,0
    orh     r11,0x0780      
2:  sub     r11,1
    breq    write_error
    ld.w    r9,r6[AVR32_MCI_SR]             // wait for ready status
    bld     r9,2
    brne    2b
    ld.w    r9,r8++
    st.w    r6[AVR32_MCI_TDR],r9
    sub     r10,1
    brne    1b

    /* wait for transfer done */
    mov     r11,0
    orh     r11,0x0780      
1:  sub     r11,1
    breq    write_error
    ld.w    r9,r6[AVR32_MCI_SR]             // wait for ready status
    bld     r9,4
    breq    1b

    /* Stop transmission */
    mov     r11,1000                        // retry 1000 times before assuming failure
1:  sub     r11,1
    breq    write_error
    ld.w    r9,r6[AVR32_MCI_SR]
    bld     r9,0
    brne    1b
    mov     r11,0
    com     r11
    mov     r10,0x2104C
    rcall   mci_cmd

    ldm     sp++,r0-r6,r9-r11,pc
write_error:
    SERIALPRINTF("Could not write SD card\r\n")
    ldm     sp++,r0-r6,r9-r11,pc
.mciaddr1: .int AVR32_MCI_ADDRESS

/**************************************************
** mci_read: read 1 sector
** r8 = buffer to read
** r7 = sector number
***************************************************/
mci_readpage:
    stm     --sp,r0-r6,r7-r11,lr

    /* Check if card inserted and initialized. */
    rcall   sd_init
    cp      r12,0
    brne    read_error1

    lsl     r7,9                            // multiply by 512, for non-HC cards, block is address in bytes

    /* wait for ready status */
    lddpc   r6,.mciaddr2
    mov     r11,0
    orh     r11,0x0780                           // retry 5sec before assuming failure
1:  sub     r11,1
    breq    read_error2
    ld.w    r10,r6[AVR32_MCI_SR]
    bld     r10,5
    brne    1b  
  
    /* CMD13 */
    lddpc   r10,.mmc_rca1
    ld.w    r11,r10[0]
    mov     r10,0x104D
    rcall   mci_cmd

    /* Set block count (1 block of 512 bytes) */
    mov     r9,1
    orh     r9,512
    lddpc   r6,.mciaddr2
    st.w    r6[AVR32_MCI_BLKR],r9 
    mov     r10,0xD1052         // multiblock read
    mov     r11,r7
    rcall   mci_cmd

    /* read the data */
    mov     r10,128                              // 128 31bit words
1:  mov     r11,0
    orh     r11,0x0780                           // retry before assuming failure
2:  sub     r11,1
    breq    read_error3
    ld.w    r9,r6[AVR32_MCI_SR]                 // wait for ready status
    bld     r9,1
    brne    2b
    ld.w    r9,r6[AVR32_MCI_RDR]
    st.w    r8++,r9
    sub     r10,1
    brne    1b

    /* Stop transmission */
    mov     r11,1000                            // retry 1000 times before assuming failure
1:  sub     r11,1
    breq    read_error4
    ld.w    r9,r6[AVR32_MCI_SR]
    bld     r9,0
    brne    1b
    mov     r11,0
    com     r11
    mov     r10,0x2104C
    rcall   mci_cmd

    ldm     sp++,r0-r6,r7-r11,pc
read_error1:
    SERIALPRINTF("Could not read SD card1\r\n")
    rjmp 1f
read_error2:
    SERIALPRINTF("Could not read SD card2\r\n")
    rjmp 1f
read_error3:
    SERIALPRINTF("Could not read SD card3\r\n")
    rjmp 1f

read_error4:
    SERIALPRINTF("Could not read SD card4\r\n")
    rjmp 1f
1:
    ldm     sp++,r0-r6,r7-r11,pc
.mmc_rca1: .int 0x04001100
.mciaddr2: .int AVR32_MCI_ADDRESS

/**************************************************************************
** sd_init
**
**************************************************************************/
mci_init:
    stm     --sp,r8-r9,lr

    /* setup GPIO */
    lddpc   r9,.mciPins
    lddpc   r8,.pioa
    st.w    r8[AVR32_PIO_PDR],r9            // disable GPIO control
    st.w    r8[AVR32_PIO_ASR],r9            // peripheral A
    st.w    r8[AVR32_PIO_PUDR],r9           // pull-up disable
    lddpc   r9,.mciPins2
    lddpc   r8,.piob
    st.w    r8[AVR32_PIO_PDR],r9            // disable GPIO control
    st.w    r8[AVR32_PIO_BSR],r9            // peripheral A
    st.w    r8[AVR32_PIO_PUDR],r9           // pull-up disable

    /* enable interrupts on PC25 change (SD card detect) */
    mov     r9,0
    sbr     r9,25
    lddpc   r8,.pioc
    st.w    r8[AVR32_PIO_PER],r9
    st.w    r8[AVR32_PIO_ODR],r9
    st.w    r8[AVR32_PIO_IFER],r9
    st.w    r8[AVR32_PIO_IER],r9
    st.w    r8[AVR32_PIO_IMR],r9

    /* clear the RCA */
    lddpc   r8,.mmc_rca2
    mov     r9,0
    st.w    r8,r9

    ldm     sp++,r8-r9,pc
.mmc_rca2: .int 0x04001100
.pioa: .int AVR32_PIOA_ADDRESS
.piob: .int AVR32_PIOB_ADDRESS
.pioc: .int AVR32_PIOC_ADDRESS
.mciPins: .int  ((1<<AVR32_MCI_CLK_0_PIN)|\
                (1<<AVR32_MCI_CMD_0_PIN)|\
                (1<<AVR32_MCI_DATA_0_PIN)|\
                (1<<AVR32_MCI_DATA_1_PIN)|\
                (1<<AVR32_MCI_DATA_2_PIN)|\
                (1<<AVR32_MCI_DATA_3_PIN))
.mciPins2: .int (1<<6)



/**************************************************************************
** sd_init
**
**************************************************************************/
sd_init:
    stm     --sp,r0-r3,r8-r11,lr

    // if RCA is set, the card is already initialized
    lddpc   r8,.mmc_rca3
    ld.w    r9,r8
    cp      r9,0
    brne    sdInitOK

    /* init MCI */
    lddpc   r8,.mciaddr
    mov     r9,1<<AVR32_MCI_CR_SWRST_OFFSET
    st.w    r8[AVR32_MCI_CR],r9             // reset
    mov     r9,1<<AVR32_MCI_CR_MCIEN_OFFSET
    st.w    r8[AVR32_MCI_CR],r9             // enable
    mov     r9,0b01011111                   
    st.w    r8[AVR32_MCI_DTOR],r9           // set timeout
    mov     r9,0
    st.w    r8[AVR32_MCI_SDCR],r9

    /* setup speed, 400khz for initialization  */
    mov     r9,0x185E                       // clkdiv = 0. PBB=CPU/2 = 75MHZ. So divide MCI clock by 0x5E*2
    orh     r9,512                          // blocksize
    st.w    r8[AVR32_MCI_MR],r9


    /* Go IDLE */
    mov     r10,0                           // GO_IDLE_STATE
    mov     r11,0
    rcall   mci_cmd

    mov     r10,0x0100                      // 74-cycle init sequence
    mov     r11,0
    rcall   mci_cmd

    /* ACMD41 */
    mov     r9,1000                     // retyrn 1000 times before timeout
1:  sub     r9,1
    brne    2f
    SERIALPRINTF("No SD card inserted\r\n")
    rjmp    sdInitError
2:  mov     r10,0x1077                 // APP cmd, maxlat=1
    mov     r11,0
    rcall   mci_cmd
    mov     r10,0x0069
    mov     r11,0
    sbr     r11,23
    sbr     r11,30
    rcall   mci_cmd
    bld     r0,31                       // send until not busy
    brne    1b

    /* complain if detected HC card. We don't support this */
    bld     r0,30
    brne    2f

    SERIALPRINTF("SDHC not supported\r\n")
    rjmp    sdInitError
2:   
 
    /* request CID */
    mov     r10,0x0082                    // MMC_CMD_ALL_SEND_CI
    mov     r11,0x00
    rcall   mci_cmd

    mov     r10,0x1043                      // SD_CMD_SEND_RELATIVE_ADDR 
    mov     r11,0
    rcall   mci_cmd                         //r0 will contain RCA
    andl    r0,0                            // clear low 16bits    
    
    /* Save rca to use it later */
    lddpc   r10,.mmc_rca3
    st.w    r10[0],r0

    mov     r10,0x1089                      // MMC_CMD_SEND_CSD
    mov     r11,r0
    rcall   mci_cmd                         // will return csd
    
    /* Set data timeout */
    lddpc   r8,.mciaddr
    mov     r9,0b01111111                   
    st.w    r8[AVR32_MCI_DTOR],r9           


    /* select card */
    lddpc   r10,.mmc_rca3
    ld.w    r11,r10[0]
    mov     r10,0x1047                      // SEL_DESEL_CARD
    rcall   mci_cmd                         

    mov     r11,1000                        // retry 1000 times before assuming failure
1:  sub     r11,1
    breq    sdInitError
    ld.w    r10,r8[AVR32_MCI_SR]    
    bld     r10,5
    brne    1b

    /* Set block size to 512 */   
    mov     r11,512
    mov     r10,0x1050                      // SEL_DESEL_CARD
    rcall   mci_cmd
    //TODO: should check if card accepts

 
    /* setup speed, 9.375MHZ  */
    mov     r9,0x1804                       // clkdiv = 0. PBB=CPU/2 = 75MHZ. So divide MCI clock by 8
    orh     r9,512                          // blocksize
    st.w    r8[AVR32_MCI_MR],r9


sdInitOK:
    mov     r12,0
    ldm     sp++,r0-r3,r8-r11,pc
sdInitError:
    mov     r12,1
    ldm     sp++,r0-r3,r8-r11,pc

.mciaddr: .int AVR32_MCI_ADDRESS
.mmc_rca3: .int 0x04001100

/***********************************
** r10 cmd (must contain response type and other flags)
** r11 arg
** r0,r1,r2,r3: response
************************************/
mci_cmd:
    stm     --sp,r8,r9,lr

    /* send command */
    lddpc   r8,.mciaddr
    st.w    r8[AVR32_MCI_ARGR],r11
    st.w    r8[AVR32_MCI_CMDR],r10

    /* wait for ready status */
    mov     r11,0x0780                        // retry  before assuming failure
    lsl     r11,16
1:  sub     r11,1
    breq    2f
    ld.w    r9,r8[AVR32_MCI_SR] 
    bld     r9,0
    brne    1b   

    /* get reply */
    bfextu  r10,r10,6,2
    breq    2f                          // if both bits are cleared, no response is expected
    ld.w    r0,r8[AVR32_MCI_RSPR]
    bld     r10,1                       // 136-bit response?
    brne    2f                          //  if so, expect 3 other words
    ld.w    r1,r8[AVR32_MCI_RSPR]
    ld.w    r2,r8[AVR32_MCI_RSPR]
    ld.w    r3,r8[AVR32_MCI_RSPR]
2:
    ldm     sp++,r8,r9,pc


