#include <avr32/io.h>
#include "macro.h"




.global LoadProcessFromELF
.global LoadProcessFromFS
.global LoadProcessFromMem
.global init_process

//////////////////////////////////////////////////////////////////////
/// init process information in kernel memory
//////////////////////////////////////////////////////////////////////
init_process:
    stm     --sp,r7,r8,r9,lr

    mov     r9,0x0400
    mov     r8,0x1000
    or      r8,r8,r9<<16
    mov     r9,0
    mov     r7,MAX_PROCESS_COUNT           // 32 entries only
1:  st.w    r8++,r9
    sub     r7,1
    brne    1b

    ldm     sp++,r7,r8,r9,pc



//////////////////////////////////////////////////////////////////////
/// Loads program in sdram and instruct scheduler to give 
///  a time share to the newly created process
//////////////////////////////////////////////////////////////////////
LoadProcessFromFS:
    stm     --sp,lr

    ldm     sp++,pc


//////////////////////////////////////////////////////////////////////
/// Loads program in sdram
///  r8=image start
///  r9=image end
//////////////////////////////////////////////////////////////////////
LoadProcessFromMem:
    stm     --sp,r0,r1,r6,r7,lr

    /* Search through process table for a free entry */
    mov     r6,0x0400
    mov     r7,0x1004                       // start at ASID 1
    or      r7,r7,r6<<16
    mov     r0,1                            // start at ASID 1
1:  ld.w    r6,r7++
    cp      r6,0
    breq    asidFree
    sub     r0,-1
    cp      r0,MAX_PROCESS_COUNT
    brlo    1b
    SERIALPRINTF("No ASID available")
    rjmp    noASIDAvailable


asidFree:
    mov     r6,0x0400
    mov     r7,0x1000
    or      r7,r7,r6<<16
    add     r7,r7,r0<<2             // r7 = 0x04001000 + (r0<<2) = address of entry in proc table
    mov     r1,r7                   // keep slot to enable later when all structures are initialized

    mov     r6,r9
    sub     r6,r8
    sub     r6,-(TASK_OVERHEAD_SIZE+4)
    lsr     r6,12
    sub     r6,-1                   // r6 contains the number of 4k pages needed for the image
    mov     r7,r8

1:  mov     r8,r0                   // allocate all pages for process
    rcall   PageAlloc
    bld     r12,31                  // if bit31 is set, this is definitaly a bad page
    breq    NoMemLeft
    sub     r6,1
    brne    1b


    /* Change ASID */
    mtsr    AVR32_TLBEHI,r0

    /* copy program */
    mov     r8,r7
    mov     r7,TASK_OVERHEAD_SIZE   // start after overhead
1:  ld.w    r6,r8++
    st.w    r7++,r6
    cp      r8,r9
    brlo    1b

    /* Prepare task context for initial values for running the first time*/
    lddpc   r8,.taskcontextbase
    mov     r7,TASK_OVERHEAD_SIZE   // start after overhead
    st.w    r8[TASKCONTEXT_RAR],r7
    mov     r7,0x100
    st.w    r8[TASKCONTEXT_SP],r7
    mov     r7,0
    st.w    r8[TASKCONTEXT_SR],r7   // mode=Application, nothing masked

    /* set back kernel ASID */
    mov     r7,0
    mtsr    AVR32_TLBEHI,r7

    /* enable task */
    mov     r7,1
    st.w    r1[0],r7

    // Task is initialized, the scheduler will execute it next time.

    rjmp    exitLoadProcessFromMem

NoMemLeft:
    SERIALPRINTF("Not enough memory")
    mov     r8,r0
    rcall   PageDeallocAll

noASIDAvailable:
exitLoadProcessFromMem:
    ldm     sp++,r0,r1,r6,r7,pc
.taskcontextbase: .int TASKCONTEXT_BASE


/**********************************************************************************
**
** r8 = adress of ELF image
**********************************************************************************/
LoadProcessFromELF:
    stm     --sp,r0-r9,lr

    /* Search through process table for a free entry */
    mov     r7,LO(0x04001004)
    orh     r7,HI(0x04001004)               // start at ASID 1
    mov     r0,1                            // start at ASID 1
1:  ld.w    r6,r7++
    cp      r6,0
    breq    asidFreeELF
    sub     r0,-1
    cp      r0,MAX_PROCESS_COUNT
    brlo    1b
    SERIALPRINTF("No ASID available")
    rjmp    LoadProcessFromELF_exit

    /* 1 free ASID found. find the slot in the table and save it for later */
asidFreeELF:
    mov     r7,LO(0x04001000)
    orh     r7,HI(0x04001000)
    add     r7,r7,r0<<2             // r7 = 0x04001000 + (r0<<2) = address of entry in proc table
    mov     r1,r7                   // r1= keep slot to enable later when all structures are initialized

    /* find 1st program header and entry point */
    ld.w    r3,r8[28]                               // program headers
    ld.w    r4,r8[24]                               // entry point
    add     r3,r3,r8                                // r3=program header base

    /* Change ASID */
    mtsr    AVR32_TLBEHI,r0

    /* find code offs and size */
    ld.w    r2,r3[4]                                // code offset
    add     r2,r2,r8
    ld.w    r10,r3[20]                               // code size
    sub     r10,-TASK_OVERHEAD_SIZE                  // take overhead into account
    lsr     r10,12
    sub     r10,-1                                   // get number of 4k pages for size

    /* reserve pages for code */
    mov     r8,r0
    rcall   PageAllocCode
    cp      r12,0
    brlt    LoadProcessFromELF_exit_err
    
    /* copy code in sdram (skip TASK_OVERHEAD) */
    sub     r12,-0x200                                  // skip TASK_OVERHEAD
    lsl     r10,10                                      // *4096 div by 4 (total number of words)
1:  ld.w    r7,r2++
    st.w    r12++,r7
    sub     r10,1
    brne    1b

    /* find data offs and size */
    ld.w    r2,r3[36]                                // data offset
    add     r2,r2,r8
    ld.w    r10,r3[52]                                // data size
    lsr     r10,12
    sub     r10,-1                                   // get number of 4k pages for size

    // create data pages
    mov     r8,r0
    rcall   PageAllocData
    cp      r12,0
    brlt    LoadProcessFromELF_exit_err

    /* copy data segment in sdram */
    lsl     r10,10                                      // *4096 div by 4 (total number of words)
1:  ld.w    r7,r2++
    st.w    r12++,r7
    sub     r10,1
    brne    1b

    /* Prepare task context for initial values for running the first time*/
    lddpc   r8,.taskcontextbase
    st.w    r8[TASKCONTEXT_RAR],r4                  // entry point as read from ELF header
    mov     r7,0x100
    st.w    r8[TASKCONTEXT_SP],r7
    mov     r7,0
    st.w    r8[TASKCONTEXT_SR],r7   // mode=Application, nothing masked

    /* set back kernel ASID */
    mov     r7,0
    mtsr    AVR32_TLBEHI,r7

    /* enable task */
    mov     r7,1
    st.w    r1[0],r7

    rjmp    LoadProcessFromELF_exit

LoadProcessFromELF_exit_err:
    SERIALPRINTF("NOT ENOUGH MEMORY\r\n")

LoadProcessFromELF_exit:
    ldm     sp++,r0-r9,pc

