#include <avr32/io.h>
#include "macro.h"
#include "vfs.h"

.global fopen
.global fread
.global fclose
.global fls

/************************************************************************
** Device drivers interface must provide the following functions
** and parameters:
** read: r8=buffer, r7=block index, r5=set to zero by calling function
** write: r8=buffer, r7=block index, r5=set to zero by calling function
************************************************************************/
deviceInterfaceMap:
    DECLARE_DEVICE_INTERFACE(dataflash_readpage,dataflash_writepage,1056,DATAFLASH_TEMPBUFFER)
    DECLARE_DEVICE_INTERFACE(mci_readpage,mci_writepage,512,MCI_TEMPBUFFER)

/*************************************************************
** This is what determines what devices use what filesystem
*************************************************************/
filesystemmaps:
    /* device 0 (/0): dataflash with patfs */ 
    DECLARE_FILESYSTEM((deviceInterfaceMap+(VFS_DEV_STRUCT_SIZE*0)),patfs_open,patfs_read,patfs_close,patfs_ls)

    /* device 1 (/1): MMC with Fat16 */
    DECLARE_FILESYSTEM((deviceInterfaceMap+(VFS_DEV_STRUCT_SIZE*1)),fat16_open,fat16_read,fat16_close,fat16_ls)









/************************************************************************
*** r7 = path
*** r9 = entry to start at
*** r6 = Max number of entries (12bytes per entry)
*** r8 = buffer to receive entries
*** RETURN r12 = number of entries read
************************************************************************/
fls:
    stm     --sp,r0-r4,lr

    mov     r4,LO(filesystemmaps)
    orh     r4,HI(filesystemmaps)

    /* get device number */
    ld.w    r0,r7
    lsr     r0,8
    mov     r1,0x302F                       // "/0/"
    mov     r2,0x312F                       // "/1/"
    orh     r1,0x2F
    orh     r2,0x2F
    cp      r0,r1
    breq    fls_device0
    cp      r0,r2
    breq    fls_device1
    mov     r12,0                               // Device not found
    SERIALPRINTF("INVALID PATH\r\n")
    rjmp    fls_exit
fls_device1:
    sub     r4,-(VFS_FS_STRUCT_SIZE*1)
    rjmp    fls_common
fls_device0:
    sub     r4,-(VFS_FS_STRUCT_SIZE*0)
    rjmp    fls_common
fls_common:
    ld.w    r3,r4[VFS_FS_DEVICESTRUCT]
    mcall   r4[VFS_FS_LS]

mov r6,r12
1: sub r6,1
breq 2f
mov r5,14
mov r7,1
rcall serialSendBuf
SERIALPRINTF("\r\n")
rjmp 1b
2:
fls_exit:
    ldm     sp++,r0-r4,pc



/************************************************************************
*** r5 = pointer to a FILE structure
************************************************************************/
fclose:
    stm     --sp,r3,r4,lr
    ld.w    r4,r5[VFS_FILE_FSTYPE]
    cp      r4,0
    breq    fclose_exit

    ld.w    r3,r4[VFS_FS_DEVICESTRUCT]
    mcall   r4[VFS_FS_CLOSE]

    mov     r4,0
    st.w    r5[VFS_FILE_FSTYPE],r4
fclose_exit:
    ldm     sp++,r3,r4,pc


/************************************************************************
*** r5 = pointer to a FILE structure
*** r11 = number of bytes to read
*** r8 = buffer which the data will be copied into
*** RETURN: r12 = number of bytes read
************************************************************************/
fread:
    stm     --sp,r3,r4,lr
    mov     r12,0
    ld.w    r4,r5[VFS_FILE_FSTYPE]
    cp      r4,0
    breq    fread_exit

    ld.w    r3,r4[VFS_FS_DEVICESTRUCT]
    mcall   r4[VFS_FS_READ]

fread_exit:
    ldm     sp++,r3,r4,pc

/************************************************************************
*** r8 = pointer to a FILE structure
*** r7 = pointer to the fully qualified name of the file
*** r12 = 0 if success, non-zero if failure
************************************************************************/
fopen:
    stm     --sp,r0-r4,lr

    mov     r4,0
    st.w    r8[VFS_FILE_FSTYPE],r4
    /* load device virtual table pointer in r3 */
    mov     r4,LO(filesystemmaps)
    orh     r4,HI(filesystemmaps)

    /* get device number */
    ld.w    r0,r7
    lsr     r0,8
    mov     r1,0x302F                       // "/0/"
    mov     r2,0x312F                       // "/1/"
    orh     r1,0x2F
    orh     r2,0x2F
    cp      r0,r1
    breq    fopen_device0
    cp      r0,r2
    breq    fopen_device1
    mov     r12,255                               // Device not found
    rjmp    fopen_exit

fopen_device1:
    sub     r4,-(VFS_FS_STRUCT_SIZE*1)
    rjmp    fopen_common
fopen_device0:
    sub     r4,-(VFS_FS_STRUCT_SIZE*0)
    rjmp    fopen_common
fopen_common:
    st.w    r8[VFS_FILE_FSTYPE],r4
    //r7 already contains fully qualified name of file
    ld.w    r3,r4[VFS_FS_DEVICESTRUCT]
    mov     r5,r8
    mcall   r4[VFS_FS_OPEN]                     // call xxx_getFirstBlock, r3 = device virtual table pointer
    cp      r12,0
    breq    fopen_exit

    mov     r4,0
    st.w    r8[VFS_FILE_FSTYPE],r4
fopen_exit:
    ldm     sp++,r0-r4,pc

